# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from taipy.core.cycle._cycle_repository_factory import _CycleRepositoryFactory
from taipy.core.data._data_repository_factory import _DataRepositoryFactory
from taipy.core.pipeline._pipeline_repository_factory import _PipelineRepositoryFactory
from taipy.core.scenario._scenario_repository_factory import _ScenarioRepositoryFactory
from taipy.core.task._task_repository_factory import _TaskRepositoryFactory

repositories = {
    "scenario": _ScenarioRepositoryFactory._build_repository,
    "pipeline": _PipelineRepositoryFactory._build_repository,
    "task": _TaskRepositoryFactory._build_repository,
    "data": _DataRepositoryFactory._build_repository,
    "cycle": _CycleRepositoryFactory._build_repository,
}


def _to_model(repository, entity, **kwargs):
    return repositories[repository](**kwargs)._to_model(entity)



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from unittest import mock

import pytest
from flask import url_for


def test_get_datanode(client, default_datanode):
    # test 404
    user_url = url_for("api.datanode_by_id", datanode_id="foo")
    rep = client.get(user_url)
    assert rep.status_code == 404

    with mock.patch("taipy.core.data._data_manager._DataManager._get") as manager_mock:
        manager_mock.return_value = default_datanode
        # test get_datanode
        rep = client.get(url_for("api.datanode_by_id", datanode_id="foo"))
        assert rep.status_code == 200


def test_delete_datanode(client):
    # test 404
    user_url = url_for("api.datanode_by_id", datanode_id="foo")
    rep = client.get(user_url)
    assert rep.status_code == 404

    with mock.patch("taipy.core.data._data_manager._DataManager._delete"), mock.patch(
        "taipy.core.data._data_manager._DataManager._get"
    ):
        # test get_datanode
        rep = client.delete(url_for("api.datanode_by_id", datanode_id="foo"))
        assert rep.status_code == 200


def test_create_datanode(client, default_datanode_config):
    # without config param
    datanodes_url = url_for("api.datanodes")
    rep = client.post(datanodes_url)
    assert rep.status_code == 400

    # config does not exist
    datanodes_url = url_for("api.datanodes", config_id="foo")
    rep = client.post(datanodes_url)
    assert rep.status_code == 404

    with mock.patch("src.taipy.rest.api.resources.datanode.DataNodeList.fetch_config") as config_mock:
        config_mock.return_value = default_datanode_config
        datanodes_url = url_for("api.datanodes", config_id="bar")
        rep = client.post(datanodes_url)
        assert rep.status_code == 201


def test_get_all_datanodes(client, default_datanode_config_list):
    for ds in range(10):
        with mock.patch("src.taipy.rest.api.resources.datanode.DataNodeList.fetch_config") as config_mock:
            config_mock.return_value = default_datanode_config_list[ds]
            datanodes_url = url_for("api.datanodes", config_id=config_mock.name)
            client.post(datanodes_url)

    rep = client.get(datanodes_url)
    assert rep.status_code == 200

    results = rep.get_json()
    assert len(results) == 10


def test_read_datanode(client, default_df_datanode):
    with mock.patch("taipy.core.data._data_manager._DataManager._get") as config_mock:
        config_mock.return_value = default_df_datanode

        # without operators
        datanodes_url = url_for("api.datanode_reader", datanode_id="foo")
        rep = client.get(datanodes_url, json={})
        assert rep.status_code == 200

        # Without operators and body
        rep = client.get(datanodes_url)
        assert rep.status_code == 200

        # TODO: Revisit filter test
        # operators = {"operators": [{"key": "a", "value": 5, "operator": "LESS_THAN"}]}
        # rep = client.get(datanodes_url, json=operators)
        # assert rep.status_code == 200


def test_write_datanode(client, default_datanode):
    with mock.patch("taipy.core.data._data_manager._DataManager._get") as config_mock:
        config_mock.return_value = default_datanode
        # Get DataNode
        datanodes_read_url = url_for("api.datanode_reader", datanode_id=default_datanode.id)
        rep = client.get(datanodes_read_url, json={})
        assert rep.status_code == 200
        assert rep.json == {"data": [1, 2, 3, 4, 5, 6]}

        datanodes_write_url = url_for("api.datanode_writer", datanode_id=default_datanode.id)
        rep = client.put(datanodes_write_url, json=[1, 2, 3])
        assert rep.status_code == 200

        rep = client.get(datanodes_read_url, json={})
        assert rep.status_code == 200
        assert rep.json == {"data": [1, 2, 3]}



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import uuid
from typing import Optional

from flask import request
from flask_restful import Resource

from taipy.config.config import Config
from taipy.core import Job
from taipy.core.common.alias import JobId
from taipy.core.exceptions.exceptions import NonExistingJob, NonExistingTaskConfig
from taipy.core.job._job_manager_factory import _JobManagerFactory
from taipy.core.task._task_manager_factory import _TaskManagerFactory

from ..exceptions.exceptions import ConfigIdMissingException
from ..middlewares._middleware import _middleware
from ..schemas import JobSchema


def _get_or_raise(job_id: str):
    manager = _JobManagerFactory._build_manager()
    job = manager._get(job_id)
    if job is None:
        raise NonExistingJob(job_id)
    return job


class JobResource(Resource):
    """Single object resource

    ---
    get:
      tags:
        - api
      summary: Get a job.
      description: |
        Return a single job by *job_id*. If the job does not exist, a 404 error is returned.

        !!! Note
          When the authorization feature is activated (available in the **Enterprise** edition only), the
          endpoint requires `TAIPY_READER` role.

        Code example:

        ```shell
          curl -X GET http://localhost:5000/api/v1/jobs/JOB_my_task_config_75750ed8-4e09-4e00-958d-e352ee426cc9
        ```

      parameters:
        - in: path
          name: job_id
          schema:
            type: string
          description: The identifier of the job.
      responses:
        200:
          content:
            application/json:
              schema:
                type: object
                properties:
                  job: JobSchema
        404:
          description: No job has the *job_id* identifier.
    delete:
      tags:
        - api
      summary: Delete a job.
      description: |
        Delete a job. If the job does not exist, a 404 error is returned.

        !!! Note
          When the authorization feature is activated (available in the **Enterprise** edition only), the endpoint
          requires `TAIPY_EDITOR` role.

        Code example:

        ```shell
          curl -X DELETE http://localhost:5000/api/v1/jobs/JOB_my_task_config_75750ed8-4e09-4e00-958d-e352ee426cc9
        ```

      parameters:
        - in: path
          name: job_id
          schema:
            type: string
          description: The identifier of the job.
      responses:
        200:
          content:
            application/json:
              schema:
                type: object
                properties:
                  message:
                    type: string
                    description: Status message.
        404:
          description: No job has the *job_id* identifier.
    """

    def __init__(self, **kwargs):
        self.logger = kwargs.get("logger")

    @_middleware
    def get(self, job_id):
        schema = JobSchema()
        job = _get_or_raise(job_id)
        return {"job": schema.dump(job)}

    @_middleware
    def delete(self, job_id):
        manager = _JobManagerFactory._build_manager()
        job = _get_or_raise(job_id)
        manager._delete(job)
        return {"message": f"Job {job_id} was deleted."}


class JobList(Resource):
    """Creation and get_all

    ---
    get:
      tags:
        - api
      summary: Get all jobs.
      description: |
        Return an array of all jobs.

        !!! Note
          When the authorization feature is activated (available in the **Enterprise** edition only), the endpoint
          requires `TAIPY_READER` role.

        Code example:

        ```shell
          curl -X GET http://localhost:5000/api/v1/jobs
        ```

      responses:
        200:
          content:
            application/json:
              schema:
                allOf:
                  - type: object
                    properties:
                      results:
                        type: array
                        items:
                          $ref: '#/components/schemas/JobSchema'
    post:
      tags:
        - api
      summary: Create a job.
      description: |
        Create a job from a task *config_id*. If the config does not exist, a 404 error is returned.

        !!! Note
          When the authorization feature is activated (available in the **Enterprise** edition only), the endpoint
          requires `TAIPY_EDITOR` role.

        Code example:

        ```shell
          curl -X POST http://localhost:5000/api/v1/jobs?task_id=TASK_my_config_75750ed8-4e09-4e00-958d-e352ee426cc9
        ```

      parameters:
        - in: query
          name: task_id
          schema:
            type: string
          description: The identifier of the task configuration.
      responses:
        201:
          content:
            application/json:
              schema:
                type: object
                properties:
                  message:
                    type: string
                    description: Status message.
                  job: JobSchema
    """

    def __init__(self, **kwargs):
        self.logger = kwargs.get("logger")

    def fetch_config(self, config_id):
        config = Config.tasks.get(config_id)
        if not config:
            raise NonExistingTaskConfig(config_id)
        return config

    @_middleware
    def get(self):
        schema = JobSchema(many=True)
        manager = _JobManagerFactory._build_manager()
        jobs = manager._get_all()
        return schema.dump(jobs)

    @_middleware
    def post(self):
        args = request.args
        task_config_id = args.get("task_id")

        if not task_config_id:
            raise ConfigIdMissingException

        manager = _JobManagerFactory._build_manager()
        schema = JobSchema()
        job = self.__create_job_from_schema(task_config_id)
        manager._set(job)
        return {
            "message": "Job was created.",
            "job": schema.dump(job),
        }, 201

    def __create_job_from_schema(self, task_config_id: str) -> Optional[Job]:
        task_manager = _TaskManagerFactory._build_manager()
        task = task_manager._bulk_get_or_create([self.fetch_config(task_config_id)])[0]
        return Job(id=JobId(f"JOB_{uuid.uuid4()}"), task=task, submit_id=f"SUBMISSION_{uuid.uuid4()}")


class JobExecutor(Resource):
    """Cancel a job

    ---
    post:
      tags:
        - api
      summary: Cancel a job.
      description: |
        Cancel a job by *job_id*. If the job does not exist, a 404 error is returned.

        !!! Note
          When the authorization feature is activated (available in the **Enterprise** edition only), the endpoint
          requires `TAIPY_EXECUTOR` role.

        Code example:

        ```shell
          curl -X POST http://localhost:5000/api/v1/jobs/cancel/JOB_my_task_config_75750ed8-4e09-4e00-958d-e352ee426cc9
        ```

      parameters:
        - in: path
          name: job_id
          schema:
            type: string
      responses:
        204:
          content:
            application/json:
              schema:
                type: object
                properties:
                  message:
                    type: string
                    description: Status message.
                  job: JobSchema
        404:
          description: No job has the *job_id* identifier.
    """

    def __init__(self, **kwargs):
        self.logger = kwargs.get("logger")

    @_middleware
    def post(self, job_id):
        manager = _JobManagerFactory._build_manager()
        job = _get_or_raise(job_id)
        manager._cancel(job)
        return {"message": f"Job {job_id} was cancelled."}



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from marshmallow import Schema, fields


class PipelineSchema(Schema):
    config_id = fields.String()
    owner_id = fields.String()
    parent_ids = fields.List(fields.String)
    tasks = fields.List(fields.String)
    version = fields.String()
    properties = fields.Dict()


class PipelineResponseSchema(PipelineSchema):
    id = fields.String()
    subscribers = fields.List(fields.Dict)



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
import json
import os

from setuptools import find_namespace_packages, find_packages, setup

with open("README.md") as readme_file:
    readme = readme_file.read()

with open(f"src{os.sep}taipy{os.sep}rest{os.sep}version.json") as version_file:
    version = json.load(version_file)
    version_string = f'{version.get("major", 0)}.{version.get("minor", 0)}.{version.get("patch", 0)}'
    if vext := version.get("ext"):
        version_string = f"{version_string}.{vext}"

setup(
    author="Avaiga",
    name="taipy-rest",
    keywords="taipy-rest",
    python_requires=">=3.8",
    version=version_string,
    author_email="dev@taipy.io",
    packages=find_namespace_packages(where="src") + find_packages(include=["taipy", "taipy.rest"]),
    package_dir={"": "src"},
    include_package_data=True,
    long_description=readme,
    long_description_content_type="text/markdown",
    description="Library to expose taipy-core REST APIs.",
    license="Apache License 2.0",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    install_requires=[
        "flask>=2.2,<3.0",
        "flask-restful>=0.3.9,<0.4",
        "flask-migrate>=3.1,<4.0",
        "flask-jwt-extended>=4.3,<5.0",
        "flask-marshmallow>=0.14,<0.15",
        "marshmallow-sqlalchemy>=0.25,<0.29",
        "passlib>=1.7.4,<1.8",
        "apispec[yaml]>=5.1,<6.0",
        "apispec-webframeworks>=0.5.2,<0.6",
        "taipy-core>=2.1,<2.2",
    ],
)
